package main

import (
	"launchpad.net/mgo"
	"launchpad.net/mgo/bson"
	"crypto/sha256"
	"crypto/rand"
	"time"
	"bytes"
	"errors"
	"encoding/hex"
	"encoding/json"
	"net/http"
	"strings"
	"log"
	"fmt"
)

type User struct {
	Login string
	Email string
	Pass []byte
	Salt []byte
}

type Object struct {
	Name string
	Owner string
	Implements string
	Published bool
	Maintainers []string
	Data []byte
}

type Session struct {
	Usid string
	User string
	Time int64
}

func main() {
	type msg map[string]interface{}
	mongo, err := mgo.Dial("localhost")
	if err != nil {
		panic(err)
	}
	defer mongo.Close()

	mongo.SetMode(mgo.Monotonic, true)

	users    := mongo.DB("tissue").C("users")
	sessions := mongo.DB("tissue").C("sessions")
	objects  := mongo.DB("tissue").C("objects")

	login := func (login, password string) (string, error) {
		result := User{}
		users.Find(bson.M{"login": login}).One(&result)

		hash := sha256.New()
		hash.Write(result.Salt)
		hash.Write([]byte(password))
		if !bytes.Equal(result.Pass, hash.Sum(nil)){
			return "", errors.New("Wrong username / password")
		}

		seed := make([]byte, 32)
		rand.Reader.Read(seed)
		if err != nil {
			return "", err
		}
		hash.Reset()
		hash.Write([]byte(result.Login))
		hash.Write([]byte(result.Email))
		hash.Write(seed)
		usid := hex.EncodeToString(hash.Sum(nil))

		n,err := sessions.Find(bson.M{"usid": usid}).Count()
		if err != nil || n != 0 {
			return "", errors.New("Hash collision!")
		}

		err = sessions.Insert(&Session{usid, login, time.Now().Unix()})
		if err != nil {
			return "", err
		}
		return usid, nil
	}

	register := func (username, email, password string) (string,error) {
		if username == "" {
			return "",errors.New("Usernames usually contain characters")
		}
		user := User{}
		n, err := users.Find(bson.M{"login": username}).Count()
		if err != nil {
			return "",err
		} else if (n != 0) {
			return "",errors.New("Username taken")
		}
		var salt []byte
		n,err = rand.Reader.Read(salt)
		if err != nil || n != len(salt) {
			return "",errors.New("Failed to seed salt")
		}
		hash := sha256.New()
		hash.Write(salt)
		hash.Write([]byte(password))
		user.Pass = hash.Sum(nil)
		user.Login = username
		user.Email = email
		user.Salt = salt

		err = users.Insert(user)
		if err != nil {
			return "",err
		}
		return login(username,password)
	}

	validate := func(username, usid string) bool {
		result := Session{}
		sessions.Find(bson.M{"usid": usid}).One(&result)
		return username == result.User
	}

	parse := func(w http.ResponseWriter, req *http.Request) {
		tmp := strings.SplitN(req.URL.Path, "/", 3)
		what := ""
		if len(tmp) >= 3 {
			what = tmp[2]
		}
		req.ParseMultipartForm(1e6);
		switch what {
		case "login":
			username := req.Form.Get("username")
			password := req.Form.Get("password")
			usid,err := login(username,password)
			var ret []byte
			if err != nil {
				fmt.Println("Login error: "+err.Error())
				ret,_ = json.Marshal(msg{"error":err.Error()})
			} else {
				fmt.Println("Login success: "+username)
				ret,_ = json.Marshal(msg{"usid":usid})
			}
			fmt.Println(string(ret))
			w.Header().Set("Content-Type", "application/json")
			w.Write(ret)
		case "logout":
			username := req.Form.Get("username")
			usid := req.Form.Get("usid")
			if (validate(username,usid)) {
				sessions.Remove(bson.M{"usid":usid})
				ret,_ := json.Marshal(msg{"error":""})
				w.Header().Set("Content-Type", "application/json")
				w.Write(ret)
				fmt.Println("Logout success: "+username)
			} else {
				fmt.Println("Logout failure: "+username)
			}
		case "register":
			username := req.Form.Get("username")
			password := req.Form.Get("password")
			email := req.Form.Get("email")
			usid,err := register(username,email,password)
			var ret []byte
			if err != nil {
				fmt.Println("Registration error: "+err.Error())
				ret,_ = json.Marshal(msg{"error":err.Error()})
			} else {
				fmt.Println("Registration success: "+username)
				ret,_ = json.Marshal(msg{"usid":usid})
			}
			fmt.Println(string(ret))
			w.Header().Set("Content-Type", "application/json")
			w.Write(ret)
		case "get":
			obj := req.Form.Get("obj")
			fmt.Println("Getting: \""+obj+"\"")
			w.Header().Set("Content-Type", "application/json")
			res := Object{}
			objects.Find(bson.M{"name": obj}).One(&res)
			ret,_ := json.Marshal(res)
			w.Write(ret)
		case "list":
			fmt.Println("Listing all objects");
			res := make([]string,0)
			objects.Find(nil).Select(bson.M{"Name":1}).All(&res)
			ret,_ := json.Marshal(msg{"data": res,"error":""})
			fmt.Println(string(ret))
			w.Header().Set("Content-Type", "application/json")
			w.Write(ret)
		case "add":
			username    := req.Form.Get("username")
			usid        := req.Form.Get("usid")
			obj         := req.Form.Get("obj")
			data        := req.Form.Get("data")
			implements  := req.Form.Get("implements")
			published   := req.Form.Get("published") == "true"
			maintainers := req.Form["maintainers"]
			if validate(username,usid) {
				objects.Insert(msg{
					"Name":obj,
					"Owner":username,
					"Implements":implements,
					"Published":published,
					"Maintainers":maintainers,
					"Data":[]byte(data),
				})
				ret,_ := json.Marshal(nil)
				w.Header().Set("Content-Type", "application/json")
				w.Write(ret)
			} else {
				ret,_ := json.Marshal(msg{"error":"invalid session"})
				w.Header().Set("Content-Type", "application/json")
				w.Write(ret)
			}
		default:
			w.Header().Set("Content-Type", "text/plain")
			w.Write([]byte("Invalid request!"))
		}
	}

	srvMux := http.NewServeMux();
	srvMux.Handle("/static/", http.FileServer(http.Dir(".")))
	srvMux.HandleFunc("/resource/", parse)
	err = http.ListenAndServeTLS(":4443", "./cert.pem","./key.pem", srvMux)
	if err != nil {
		log.Fatal("ListenAndServe: ", err)
	}
}
