
define(function () { return function (canvas) {
	var context = canvas.getContext("2d");

	context.save();

	var setView = function(x, y, sx, sy) {
		context.restore();
		context.save();
		context.translate(-x, -y);
		context.scale(sx || 1, sy || sx || 1);
	}
	
	var drawShape = function(shape) {
		if(!shape) return;
		context.save();
		if(shape.pos) {
			context.translate(shape.pos.x, shape.pos.y);

		}
		if(shape.angle) {
			if(shape.angle.value) {
				context.rotate(shape.angle.value);
			}
			else {
				context.rotate(shape.angle);
			}
		}
		if(shape.fixedAngle) {
			if(shape.globalAngle.value) {
				context.rotate(-shape.globalAngle.value);
			}
		}

		if(Array.isArray(shape)) {
			shape.forEach(function(subShape) {
				drawShape(subShape);
			});
		}

		if(shape.shape) {
			drawShape(shape.shape);
		}

		if(shape.path) {
			path = shape.path;
			tracePath(shape.path);
			if (shape.closed){
				context.closePath();
			}
			if (shape.fillStyle){
				if(typeof shape.fillStyle == "object") {
					context.fillStyle = shape.fillStyle.value;
				} else {
					context.fillStyle = shape.fillStyle;
				}
				context.fill();
			}
			if (shape.strokeStyle){
				if(shape.lineWidth) {
					if(typeof shape.lineWidth == "object") {
						context.lineWidth = shape.lineWidth.value;
					} else {
						context.lineWidth = shape.lineWidth;
					}
				}
				if(typeof shape.strokeStyle == "object") {
					if(shape.strokeStyle.value) {
						context.strokeStyle = shape.strokeStyle.value;
						context.stroke();
					}
				} else if(typeof shape.strokeStyle == "string") {
					context.strokeStyle = shape.strokeStyle;
					context.stroke();
				}
			}
		}
		else if(shape.type == "text") {
			var alignment = shape.align;
			var baseline = shape.baseline;
			if(shape.alwaysUpright) {
				var testVal = shape.globalAngle.value-Math.PI/2;
				if(testVal < 0) testVal += 2*Math.PI;
 				if(testVal > 0 && testVal <= Math.PI) {
					context.rotate(Math.PI);
					if(alignment == "left") alignment = "right";
					else if(alignment == "right") alignment = "left";
					if(baseline == "top") baseline = "bottom";
					else if(baseline == "bottom") baseline = "top";
				}
			}
			context.font = shape.font;
			context.textBaseline = baseline;
			context.textAlign = alignment;
			if(typeof shape.text == "string") {
				context.fillText(shape.text, 0, 0);
			}
			else if(typeof shape.text == "object") {
				context.fillText(shape.text.value, 0, 0);
			}
		}

		context.restore();
	}

	var isPointInShape = function(point, shape) {
		var pointInShape = false;
		context.save();
		if(shape.pos) {
			context.translate(shape.pos.x, shape.pos.y);
		}
		if(shape.angle) {
			context.rotate(shape.angle);
		}
		if(shape.boundingPath) { 
			tracePath(shape.boundingPath);
			context.closePath();
		}
		else if(shape.path) { 
			tracePath(shape.path);
			context.closePath();
		}
		else if(shape.shape) {
			pointInShape = isPointInShape(point, shape.shape);
		}
		if(context.isPointInPath(point.x, point.y)) pointInShape = true;
		context.restore();
		return pointInShape;
	}

	var actionAreaAtPoint = function(point, shape) {
		var actionArea = null;
		context.save();
		if(Array.isArray(shape)) {
			shape.forEach(function(subshape) {
				var result = actionAreaAtPoint(point, subshape);
				if(result) actionArea = result;
			});
			context.restore();
			return actionArea;
		}
		if(shape.pos) {
			context.translate(shape.pos.x, shape.pos.y);
		}
		if(shape.angle) {
			context.rotate(shape.angle);
		}
		if(shape.actionArea) { 
			if(Array.isArray(shape.actionArea)) {
				shape.actionArea.forEach(function(area) {
					tracePath(area.path);
					context.closePath();
					if(context.isPointInPath(point.x, point.y)) actionArea = area;
				});
			}
		}
		if(shape.shape) {
			if(Array.isArray(shape.shape)) {
				shape.shape.forEach(function(subshape) {
					var result = actionAreaAtPoint(point, subshape);
					if(result) {
						actionArea = result;
					}
				});
			}
			else {
				var result = actionAreaAtPoint(point,shape.shape);
				if(result) {
					actionArea = result;
				}
			}
		}
		context.restore();

		return actionArea;
	}

	var clearCanvas = function() {
		context.save();
		context.setTransform(1, 0, 0, 1, 0, 0);
		context.clearRect(0,0,canvas.width, canvas.height);
		context.restore();
	}

	var tracePath = function(path) {
		context.beginPath();
		if(Array.isArray(path)) {
			path.forEach(function(segment) {
				switch (segment.type) {
				case "arc":
					var radius = 0;
					if(segment.radius) {
						if(segment.radius.value) radius = segment.radius.value;
						else radius = segment.radius;
					}
					if(segment.pos)
						context.arc(segment.pos.x, segment.pos.y, radius, segment.start, segment.end, segment.ccw);
					else context.arc(segment.x, segment.y, radius, segment.start, segment.end, segment.ccw);

					break;
				case "line":
					try{
						context.moveTo(segment.points[0].x, segment.points[0].y);
						segment.points.forEach(function(point){
							context.lineTo(point.x, point.y);
						});
					} catch(err) {
					}
					break;
				case "rect":
					context.rect(segment.x, segment.y, segment.w, segment.h);
					break;
				}
			});
		} else {
			console.log("path is not path!");
		}
	}

	var globalPosition = function(point) {

	}

	return {
		"drawShape": drawShape,
		"clearCanvas": clearCanvas,
		"isPointInShape": isPointInShape,
		"actionAreaAtPoint": actionAreaAtPoint,
		"tracePath": tracePath,
		"globalPosition": globalPosition
	}
}});
